"""Base classes for PyNHD functions."""

# pyright: reportGeneralTypeIssues=false
from __future__ import annotations

import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Literal, Union, cast, overload

import cytoolz.curried as tlz
import geopandas as gpd
import pandas as pd
import shapely
import ujson
from shapely import MultiPoint, MultiPolygon, Polygon
from shapely import box as shapely_box
from shapely.geometry import mapping as shapely_mapping

import async_retriever as ar
import pygeoutils as geoutils
from pygeoogc import ArcGISRESTful, ServiceURL
from pygeoogc import utils as ogc_utils
from pygeoutils import EmptyResponseError
from pynhd.exceptions import (
    InputRangeError,
    InputTypeError,
    InputValueError,
    MissingColumnError,
    MissingCRSError,
    MissingItemError,
    ServiceError,
    ZeroMatchedError,
)

if TYPE_CHECKING:
    import pyproj
    from shapely import LineString, Point

    CRSTYPE = Union[int, str, pyproj.CRS]
    GTYPE = Union[
        Polygon, MultiPolygon, MultiPoint, LineString, Point, "tuple[float, float, float, float]"
    ]

__all__ = ["AGRBase", "ScienceBase", "GeoConnex"]


def get_parquet(parquet_path: Path | str) -> Path:
    """Get a parquet filename from a path or a string."""
    if Path(parquet_path).suffix != ".parquet":
        raise InputValueError("parquet_path", ["a filename with `.parquet` extension."])

    output = Path(parquet_path)
    output.parent.mkdir(parents=True, exist_ok=True)

    return output


@dataclass(frozen=True)
class ServiceInfo:
    """Information about a web service."""

    url: str
    layer: str
    extent: tuple[float, float, float, float] | None
    feature_types: dict[int, str] | None


class AGRBase:
    """Base class for getting geospatial data from a ArcGISRESTful service.

    Parameters
    ----------
    base_url : str, optional
        The ArcGIS RESTful service url. The URL must either include a layer number
        after the last ``/`` in the url or the target layer must be passed as an argument.
    layer : str, optional
        A valid service layer. To see a list of available layers instantiate the class
        without passing any argument.
    outfields : str or list, optional
        Target field name(s), default to "*" i.e., all the fields.
    crs : str, int, or pyproj.CRS, optional
        Target spatial reference, default to ``EPSG:4326``
    outformat : str, optional
        One of the output formats offered by the selected layer. If not correct
        a list of available formats is shown, defaults to ``json``.
    """

    def __init__(
        self,
        base_url: str,
        layer: str | None = None,
        outfields: str | list[str] = "*",
        crs: CRSTYPE = 4326,
        outformat: str = "json",
    ) -> None:
        if isinstance(layer, str):
            valid_layers = self.get_validlayers(base_url)
            try:
                self.client = ArcGISRESTful(
                    base_url,
                    valid_layers[layer.lower()],
                    outformat=outformat,
                    outfields=outfields,
                    crs=crs,
                )
            except KeyError as ex:
                raise InputValueError("layer", list(valid_layers)) from ex
        else:
            self.client = ArcGISRESTful(
                base_url,
                None,
                outformat=outformat,
                outfields=outfields,
                crs=crs,
            )

        full_layer = self.client.client.valid_layers[str(self.client.client.layer)]
        self._service_info = ServiceInfo(
            self.client.client.base_url,
            f"{full_layer} ({self.client.client.layer})",
            self.client.client.extent,
            self.client.client.feature_types,
        )

    @property
    def service_info(self) -> ServiceInfo:
        """Get the service information."""
        return self._service_info

    @staticmethod
    def get_validlayers(url: str) -> dict[str, int]:
        """Get a list of valid layers.

        Parameters
        ----------
        url : str
            The URL of the ArcGIS REST service.

        Returns
        -------
        dict
            A dictionary of valid layers.
        """
        rjson = ar.retrieve_json([url], [{"params": {"f": "json"}}])
        rjson = cast("list[dict[str, Any]]", rjson)
        if not rjson[0]["layers"]:
            ar.delete_url_cache(url, params={"f": "json"})
            with ogc_utils.RetrySession(disable=True) as session:
                rjson = session.get(url, params={"f": "json"}).json()
        if not rjson[0]["layers"]:
            msg = "The service doesn't have any layers."
            raise ServiceError(msg, url)
        return {lyr["name"].lower(): int(lyr["id"]) for lyr in rjson[0]["layers"]}

    def _getfeatures(
        self, oids: Iterator[tuple[str, ...]], return_m: bool = False, return_geom: bool = True
    ) -> gpd.GeoDataFrame:
        """Send a request for getting data based on object IDs.

        Parameters
        ----------
        return_m : bool
            Whether to activate the Return M (measure) in the request, defaults to False.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        geopandas.GeoDataFrame
            The requested features as a GeoDataFrame.
        """
        try:
            return geoutils.json2geodf(
                self.client.get_features(oids, return_m, return_geom), self.client.client.crs
            )
        except EmptyResponseError as ex:
            raise ZeroMatchedError from ex

    def bygeom(
        self,
        geom: Polygon | list[tuple[float, float]] | tuple[float, float, float, float],
        geo_crs: CRSTYPE = 4326,
        sql_clause: str = "",
        distance: int | None = None,
        return_m: bool = False,
        return_geom: bool = True,
    ) -> gpd.GeoDataFrame:
        """Get feature within a geometry that can be combined with a SQL where clause.

        Parameters
        ----------
        geom : Polygon or tuple
            A geometry (Polygon) or bounding box (tuple of length 4).
        geo_crs : str
            The spatial reference of the input geometry.
        sql_clause : str, optional
            A valid SQL 92 WHERE clause, defaults to an empty string.
        distance : int, optional
            The buffer distance for the input geometries in meters, default to None.
        return_m : bool, optional
            Whether to activate the Return M (measure) in the request, defaults to False.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        geopandas.GeoDataFrame
            The requested features as a GeoDataFrame.
        """
        oids = self.client.oids_bygeom(
            geom, geo_crs=geo_crs, sql_clause=sql_clause, distance=distance
        )
        return self._getfeatures(oids, return_m, return_geom)

    def byids(
        self,
        field: str,
        fids: str | list[str],
        return_m: bool = False,
        return_geom: bool = True,
    ) -> gpd.GeoDataFrame:
        """Get features based on a list of field IDs.

        Parameters
        ----------
        field : str
            Name of the target field that IDs belong to.
        fids : str or list
            A list of target field ID(s).
        return_m : bool
            Whether to activate the Return M (measure) in the request, defaults to False.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        geopandas.GeoDataFrame
            The requested features as a GeoDataFrame.
        """
        oids = self.client.oids_byfield(field, fids)
        return self._getfeatures(oids, return_m, return_geom)

    def bysql(
        self,
        sql_clause: str,
        return_m: bool = False,
        return_geom: bool = True,
    ) -> gpd.GeoDataFrame:
        """Get feature IDs using a valid SQL 92 WHERE clause.

        Notes
        -----
        Not all web services support this type of query. For more details look
        `here <https://developers.arcgis.com/rest/services-reference/query-feature-service-.htm#ESRI_SECTION2_07DD2C5127674F6A814CE6C07D39AD46>`__

        Parameters
        ----------
        sql_clause : str
            A valid SQL 92 WHERE clause.
        return_m : bool
            Whether to activate the measure in the request, defaults to False.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        geopandas.GeoDataFrame
            The requested features as a GeoDataFrame.
        """
        oids = self.client.oids_bysql(sql_clause)
        return self._getfeatures(oids, return_m, return_geom)

    def __repr__(self) -> str:
        """Print the service configuration."""
        return self.client.__repr__()


class PyGeoAPIBase:
    """Access `PyGeoAPI <https://labs.waterdata.usgs.gov/api/nldi/pygeoapi>`__ service."""

    def __init__(self) -> None:
        self.base_url = ServiceURL().restful.pygeoapi
        self.req_idx: list[int | str] = [0]

    def get_url(self, operation: str) -> str:
        """Set the service url."""
        return f"{self.base_url}/nldi-{operation}/execution"

    @staticmethod
    def request_body(
        id_value: list[dict[str, Any]],
    ) -> list[dict[str, dict[str, list[dict[str, Any]]]]]:
        """Return a valid request body."""
        return [
            {
                "json": {
                    "inputs": [
                        {
                            "id": i,
                            "type": "text/plain",
                            "value": list(v) if isinstance(v, (list, tuple)) else str(v),
                        }
                        for i, v in iv.items()
                    ]
                }
            }
            for iv in id_value
        ]

    def get_response(
        self, url: str, payload: list[dict[str, dict[str, list[dict[str, Any]]]]]
    ) -> gpd.GeoDataFrame:
        """Post the request and return the response as a GeoDataFrame."""
        resp = ar.retrieve_json([url] * len(payload), payload, "POST")
        resp = cast("list[dict[str, Any]]", resp)
        nfeat = len(resp)
        try:
            idx, resp = zip(*((i, r) for i, r in enumerate(resp) if "code" not in r))
            idx = cast("tuple[int]", idx)
            if len(idx) != nfeat:
                _ = [
                    ar.delete_url_cache(url, **p)  # pyright: ignore[reportArgumentType]
                    for i, p in enumerate(payload)
                    if i not in idx
                ]
            resp = cast("tuple[dict[str, Any]]", resp)

            if len(resp) < nfeat:
                warnings.warn(
                    " ".join(
                        [
                            f"There are {nfeat - len(resp)} requests that",
                            "didn't return any feature. Check their parameters and retry.",
                        ]
                    ),
                    UserWarning,
                    stacklevel=2,
                )
            gdf = gpd.GeoDataFrame(  # pyright: ignore[reportCallIssue]
                pd.concat(
                    (geoutils.json2geodf(r) for r in resp), keys=[self.req_idx[i] for i in idx]
                ),
                crs=4326,
            )
        except ValueError as ex:
            msg = "Invalid inpute parameters, check them and retry."
            raise ServiceError(msg) from ex
        except EmptyResponseError as ex:
            raise ZeroMatchedError from ex
        drop_cols = ["level_1", "spatial_ref"] if "spatial_ref" in gdf else ["level_1"]
        return (  # pyright: ignore[reportReturnType]
            gdf.reset_index().rename(columns={"level_0": "req_idx"}).drop(columns=drop_cols)
        )

    @staticmethod
    def check_coords(
        coords: tuple[float, float] | list[tuple[float, float]],
        crs: CRSTYPE,
    ) -> list[tuple[float, float]]:
        """Check the coordinates."""
        try:
            mps = MultiPoint(coords)  # pyright: ignore[reportArgumentType]
        except (TypeError, AttributeError):
            try:
                mps = MultiPoint([coords])
            except ValueError as ex:
                raise InputTypeError(
                    "coords", "tuple or list of them", "(x, y) or [(x, y), ...]"
                ) from ex
        except ValueError as ex:
            raise InputTypeError(
                "coords", "tuple or list of them", "(x, y) or [(x, y), ...]"
            ) from ex

        _coords = [(p.x, p.y) for p in mps.geoms]
        _coords = [
            (round(x, 6), round(y, 6)) for x, y in geoutils.geometry_reproject(_coords, crs, 4326)
        ]
        return _coords


class PyGeoAPIBatch(PyGeoAPIBase):
    """Access `PyGeoAPI <https://labs.waterdata.usgs.gov/api/nldi/pygeoapi>`__ service.

    Parameters
    ----------
    coords : geopandas.GeoDataFrame
        A GeoDataFrame containing the coordinates to query. The indices of the
        GeoDataFrame are used as the request IDs that will be returned in the
        response in a column named ``req_idx``.
        The required columns for using the class methods are:

        * ``flow_trace``: ``direction`` that indicates the direction of the flow trace.
          It can be ``up``, ``down``, or ``none``.
        * ``split_catchment``: ``upstream`` that indicates whether to return all upstream
          catchments or just the local catchment.
        * ``elevation_profile``: ``numpts`` that indicates the number of points to extract
          along the flowpath and ``dem_res`` that indicates the target resolution for
          requesting the DEM from 3DEP service.
        * ``cross_section``: ``numpts`` that indicates the number of points to extract
          along the flowpath and ``width`` that indicates the width of the cross-section
          in meters.
    """

    def __init__(self, coords: gpd.GeoDataFrame) -> None:
        super().__init__()
        if coords.crs is None:
            raise MissingCRSError

        self.coords = cast("gpd.GeoDataFrame", coords.to_crs(4326))
        self.req_idx = self.coords.index.tolist()
        self.req_cols = {
            "flow_trace": ["direction"],
            "split_catchment": ["upstream"],
            "elevation_profile": ["numpts", "dem_res"],
            "endpoints_profile": ["numpts", "dem_res"],
            "cross_section": ["numpts", "width"],
        }
        self.geo_types = {
            "flow_trace": "Point",
            "split_catchment": "Point",
            "elevation_profile": "LineString",
            "endpoints_profile": "MultiPoint",
            "cross_section": "Point",
        }
        self.service = {
            "flow_trace": "flowtrace",
            "split_catchment": "splitcatchment",
            "elevation_profile": "xsatpathpts",
            "endpoints_profile": "xsatendpts",
            "cross_section": "xsatpoint",
        }

    def check_col(self, method: str) -> None:
        """Check if the required columns are present in the GeoDataFrame."""
        missing = [c for c in self.req_cols[method] if c not in self.coords]
        if missing:
            raise MissingColumnError(missing)

    def check_geotype(self, method: str) -> None:
        """Check if the required geometry type is present in the GeoDataFrame."""
        if not self.coords.geom_type.eq(self.geo_types[method]).all():
            raise InputTypeError("coords", self.geo_types[method])

    def get_payload(self, method: str) -> list[dict[str, dict[str, list[dict[str, Any]]]]]:
        """Return the payload for a request."""
        self.check_col(method)
        self.check_geotype(method)

        attrs = self.req_cols[method]

        if "dem_res" in self.coords:
            coords = self.coords.rename(columns={"dem_res": "3dep_res"})
            attrs = ["numpts", "3dep_res"]
        else:
            coords = self.coords

        geo_iter = coords[["geometry", *attrs]].itertuples(index=False, name=None)

        if method == "endpoints_profile":
            if any(len(g.geoms) != 2 for g in coords.geometry):
                raise InputTypeError("coords", "MultiPoint of length 2")

            return self.request_body(
                [
                    {
                        "lat": [round(g.y, 6) for g in mp.geoms],
                        "lon": [round(g.x, 6) for g in mp.geoms],
                        **dict(zip(attrs, list(u))),
                    }
                    for mp, *u in geo_iter
                ]
            )

        if method == "elevation_profile":
            return self.request_body(
                [
                    {
                        "path": [[round(x, 6), round(y, 6)] for x, y in line.coords],
                        **dict(zip(attrs, list(u))),
                    }
                    for line, *u in geo_iter
                ]
            )

        return self.request_body(
            [
                {"lat": round(g.y, 6), "lon": round(g.x, 6), **dict(zip(attrs, list(u)))}
                for g, *u in geo_iter
            ]
        )


class ScienceBase:
    """Access and explore items on USGS's ScienceBase."""

    @staticmethod
    def get_children(item: str) -> dict[str, Any]:
        """Get children items of an item."""
        url = "https://www.sciencebase.gov/catalog/items"
        payload = {
            "filter": f"parentIdExcludingLinks={item}",
            "fields": "title,id",
            "format": "json",
        }
        resp = ar.retrieve_json([url], [{"params": payload}])
        resp = cast("list[dict[str, Any]]", resp)
        return resp[0]

    @staticmethod
    def get_file_urls(item: str) -> pd.DataFrame:
        """Get download and meta URLs of all the available files for an item."""
        url = "https://www.sciencebase.gov/catalog/item"
        payload = {"fields": "files,downloadUri", "format": "json"}
        resp = ar.retrieve_json([f"{url}/{item}"], [{"params": payload}])
        resp = cast("list[dict[str, Any]]", resp)
        urls = zip(
            tlz.pluck("name", resp[0]["files"]),
            tlz.pluck("url", resp[0]["files"]),
            tlz.pluck("metadataHtmlViewUri", resp[0]["files"], default=None),
        )
        return pd.DataFrame(
            urls,
            columns=["name", "url", "metadata_url"],  # pyright: ignore[reportArgumentType]
        ).set_index("name")


@dataclass(frozen=True)
class GCXURL:
    items: str
    queryables: str


@dataclass(frozen=True)
class EndPoints:
    name: str
    description: str
    url: str
    query_fields: list[str]
    extent: tuple[float, float, float, float]
    dtypes: dict[str, str]


class GeoConnex:
    """Access to the GeoConnex API.

    Notes
    -----
    The ``geometry`` field of the query can be a Polygon, MultiPolygon,
    or tuple/list of length 4 (bbox) in ``EPSG:4326`` CRS. They should
    be within the extent of the GeoConnex endpoint.

    Parameters
    ----------
    The item (service endpoint) to query. Valid endpoints are:

        - ``hu02`` for Two-digit Hydrologic Regions
        - ``hu04`` for Four-digit Hydrologic Subregion
        - ``hu06`` for Six-digit Hydrologic Basins
        - ``hu08`` for Eight-digit Hydrologic Subbasins
        - ``hu10`` for Ten-digit Watersheds
        - ``nat_aq`` for National Aquifers of the United States from
            USGS National Water Information System National Aquifer code list.
        - ``principal_aq`` for Principal Aquifers of the United States from
            2003 USGS data release
        - ``sec_hydrg_reg`` for Secondary Hydrogeologic Regions of the
            Conterminous United States from 2018 USGS data release
        - ``gages`` for US Reference Stream Gauge Monitoring Locations
        - ``mainstems`` for US Reference Mainstem Rivers
        - ``states`` for U.S. States
        - ``counties`` for U.S. Counties
        - ``aiannh`` for Native American Lands
        - ``cbsa`` for U.S. Metropolitan and Micropolitan Statistical Areas
        - ``ua10`` for Urbanized Areas and Urban Clusters (2010 Census)
        - ``places`` for U.S. legally incororated and Census designated places
        - ``pws`` for U.S. Public Water Systems
        - ``dams`` for US Reference Dams

    dev : bool, optional
        Whether to use the development endpoint, defaults to ``False``.
    max_nfeatures : int, optional
        The maximum number of features to request from the service,
        defaults to 10000.
    """

    def __init__(
        self, item: str | None = None, dev: bool = False, max_nfeatures: int = 10000
    ) -> None:
        self.dev = dev
        self.item = item
        self.max_nfeatures = max_nfeatures

    @property
    def dev(self) -> bool:
        """Return the name of the endpoint."""
        return self._dev

    @dev.setter
    def dev(self, value: bool) -> None:
        self._dev = value
        if value:
            self.base_url = f"{ServiceURL().restful.geoconnex.replace('.us', '.dev')}/collections"
        else:
            self.base_url = f"{ServiceURL().restful.geoconnex}/collections"

        resp = ar.retrieve_json([self.base_url], [{"params": {"f": "json"}}])
        resp = cast("list[dict[str, Any]]", resp)
        pluck = tlz.partial(tlz.pluck, seqs=resp[0]["collections"])

        def get_links(links: list[dict[str, Any]]) -> dict[str, str | list[str] | dict[str, str]]:
            """Get links."""
            urls = {
                lk["rel"]: lk["href"].replace("?f=json", "")
                for lk in links
                if lk["type"] == "application/json"
            }
            resp = ar.retrieve_json([f"{urls['self']}/queryables"], [{"params": {"f": "json"}}])
            resp = cast("list[dict[str, Any]]", resp)
            prop: dict[str, dict[str, str]] = resp[0]["properties"]
            _ = prop.pop("geometry", None)
            fields = list(prop)
            type_map = {
                "integer": "int64",
                "string": "str",
                "number": "f8",
                "timestamp": "datetime64[ns]",
                "boolean": "bool",
            }
            dtypes = {v["title"]: type_map[v["type"]] for v in prop.values()}

            return {
                "url": f"{urls['self']}/items",
                "query_fields": fields,
                "dtypes": dtypes,
            }

        eps = zip(
            pluck(ind="id"), pluck(ind="description"), pluck(ind="links"), pluck(ind="extent")
        )
        self.endpoints = {
            ep[0]: EndPoints(
                name=ep[0],
                description=ep[1],
                **get_links(ep[2]),  # pyright: ignore[reportArgumentType]
                extent=tuple(ep[3]["spatial"]["bbox"][0]),  # pyright: ignore[reportArgumentType]
            )
            for ep in eps
        }

    @property
    def item(self) -> str | None:
        """Return the name of the endpoint."""
        return self._item

    @item.setter
    def item(self, value: str | None) -> None:
        self._item = value
        if value is not None:
            self.dev = self._dev
            valid = [f"- ``{i}`` for {e.description}" for i, e in self.endpoints.items()]
            if value not in self.endpoints:
                raise InputValueError("item", valid)
            self.query_url = self.endpoints[value].url
            self.item_extent = self.endpoints[value].extent
        else:
            self.query_url = None

    def _get_geodf(self, kwds: dict[str, Any]) -> gpd.GeoDataFrame | pd.DataFrame:
        """Get response as a ``geopandas.GeoDataFrame``."""
        skip_geometry = kwds.pop("skipGeometry") if "skipGeometry" in kwds else "false"
        params = {
            **kwds,
            "headers": {"Content-Type": "application/query-cql-json"},
        }

        def _get_url(offset: int) -> str:
            url_kwds = {
                "f": "json",
                "limit": self.max_nfeatures,
                "offset": offset,
                "sortby": "uri",
                "filter-lang": "cql-json",
                "skipGeometry": skip_geometry,
            }
            return f"{self.query_url}?{'&'.join([f'{k}={v}' for k, v in url_kwds.items()])}"

        resp = ar.retrieve_json([_get_url(0)], [params], request_method="POST")
        resp = cast("list[dict[str, Any]]", resp)
        if resp[0].get("code") is not None:
            msg = f"{resp[0]['code']}: {resp[0]['description']}"
            raise ServiceError(msg, _get_url(0))
        try:
            gdf = geoutils.json2geodf(resp)
            n_matched = int(resp[0]["numberMatched"])
        except EmptyResponseError as ex:
            raise ZeroMatchedError from ex

        if len(gdf) != n_matched:
            urls = [_get_url(i) for i in range(0, n_matched, self.max_nfeatures)]
            resp = ar.retrieve_json(urls, [params] * len(urls), request_method="POST")
            resp = cast("list[dict[str, Any]]", resp)
            gdf = geoutils.json2geodf(resp)

        if len(gdf) == 0:
            raise ZeroMatchedError

        if "geometry" in gdf:
            return gdf.reset_index(drop=True)
        return pd.DataFrame(gdf.reset_index(drop=True))

    @overload
    def bygeometry(
        self,
        geometry1: GTYPE,
        geometry2: GTYPE | None = ...,
        predicate: str = ...,
        crs: CRSTYPE = ...,
        skip_geometry: Literal[False] = False,
    ) -> gpd.GeoDataFrame: ...

    @overload
    def bygeometry(
        self,
        geometry1: GTYPE,
        geometry2: GTYPE | None = ...,
        predicate: str = ...,
        crs: CRSTYPE = ...,
        skip_geometry: Literal[True] = True,
    ) -> pd.DataFrame: ...

    def bygeometry(
        self,
        geometry1: GTYPE,
        geometry2: GTYPE | None = None,
        predicate: str = "intersects",
        crs: CRSTYPE = 4326,
        skip_geometry: bool = False,
    ) -> gpd.GeoDataFrame | pd.DataFrame:
        """Query the GeoConnex endpoint by geometry.

        Parameters
        ----------
        geometry1 : Polygon or tuple of float
            The first geometry or bounding boxes to query. A bounding box is
            a tuple of length 4 in the form of ``(xmin, ymin, xmax, ymax)``.
            For example, an spatial query for a single geometry would be
            ``INTERSECTS(geom, geometry1)``.
        geometry2 : Polygon or tuple of float, optional
            The second geometry or bounding boxes to query. A bounding box is
            a tuple of length 4 in the form of ``(xmin, ymin, xmax, ymax)``.
            Default is ``None``. For example, an spatial query for a two
            geometries would be ``CROSSES(geometry1, geometry2)``.
        predicate : str, optional
            The predicate to use, by default ``intersects``. Supported
            predicates are ``intersects``, ``within``, ``contains``,
            ``overlaps``, ``crosses``, ``disjoint``, ``touches``, and
            ``equals``.
        crs : int or str or pyproj.CRS, optional
            The CRS of the polygon, by default ``EPSG:4326``. If the input
            is a ``geopandas.GeoDataFrame`` or ``geopandas.GeoSeries``,
            this argument will be ignored.
        skip_geometry: bool, optional
            If ``True``, no geometry will not be returned.

        Returns
        -------
        geopandas.GeoDataFrame
            The query result as a ``geopandas.GeoDataFrame``.
        """
        if self.item is None or self.query_url is None:
            raise MissingItemError(["item"])

        valid_predicates = (
            "INTERSECTS",
            "WITHIN",
            "CONTAINS",
            "OVERLAPS",
            "CROSSES",
            "DISJOINT",
            "TOUCHES",
            "EQUALS",
        )
        if predicate.upper() not in valid_predicates:
            raise InputValueError("predicate", valid_predicates)

        geom1 = geoutils.geo2polygon(geometry1, crs, 4326)  # pyright: ignore[reportArgumentType]
        if not geom1.intersects(shapely_box(*self.item_extent)):
            raise InputRangeError("geometry", f"within {self.item_extent}")
        try:
            geom1_json = ujson.loads(shapely.to_geojson(geom1))
        except AttributeError:
            geom1_json = shapely_mapping(geom1)

        if geometry2 is None:
            return self._get_geodf(
                {
                    "json": {predicate.lower(): [{"property": "geom"}, geom1_json]},
                    "skipGeometry": str(skip_geometry).lower(),
                }
            )

        geom2 = geoutils.geo2polygon(geometry2, crs, 4326)  # pyright: ignore[reportArgumentType]
        if not geom2.intersects(shapely_box(*self.item_extent)):
            raise InputRangeError("geometry", f"within {self.item_extent}")
        try:
            geom_json2 = ujson.loads(shapely.to_geojson(geom2))
        except AttributeError:
            geom_json2 = shapely_mapping(geom2)
        return self._get_geodf(
            {
                "json": {predicate.lower(): [geom1_json, geom_json2]},
                "skipGeometry": str(skip_geometry).lower(),
            }
        )

    @overload
    def byid(
        self,
        feature_name: str,
        feature_ids: list[str] | str,
        skip_geometry: Literal[False] = False,
    ) -> gpd.GeoDataFrame: ...

    @overload
    def byid(
        self,
        feature_name: str,
        feature_ids: list[str] | str,
        skip_geometry: Literal[True],
    ) -> pd.DataFrame: ...

    def byid(
        self,
        feature_name: str,
        feature_ids: list[str] | str,
        skip_geometry: bool = False,
    ) -> gpd.GeoDataFrame | pd.DataFrame:
        """Query the GeoConnex endpoint."""
        if self.item is None or self.query_url is None:
            raise MissingItemError(["item"])

        fids = {feature_ids} if isinstance(feature_ids, (str, int)) else set(feature_ids)

        valid_keys = self.endpoints[self.item].query_fields
        if feature_name not in valid_keys:
            raise InputValueError("feature_name", valid_keys)
        ftyped = pd.Series(list(fids)).astype(self.endpoints[self.item].dtypes[feature_name])
        kwds = {
            "json": {
                "in": {
                    "value": {"property": feature_name},
                    "list": ftyped.to_list(),
                },
                "skipGeometry": str(skip_geometry).lower(),
            }
        }
        return self._get_geodf(kwds)

    @overload
    def bycql(
        self,
        cql_dict: dict[str, Any],
        skip_geometry: Literal[False] = False,
    ) -> gpd.GeoDataFrame: ...

    @overload
    def bycql(
        self,
        cql_dict: dict[str, Any],
        skip_geometry: Literal[True],
    ) -> pd.DataFrame: ...

    def bycql(
        self,
        cql_dict: dict[str, Any],
        skip_geometry: bool = False,
    ) -> gpd.GeoDataFrame | pd.DataFrame:
        """Query the GeoConnex endpoint.

        Notes
        -----
        GeoConnex only supports simple CQL queries. For more information
        and examples visit https://portal.ogc.org/files/96288#simple-cql-JSON.
        Use this for non-spatial queries, since there's a dedicated method
        for spatial queries, :meth:`.bygeometry`.

        Parameters
        ----------
        cql_dict : dict
            A valid CQL dictionary (non-spatial queries).
        skip_geometry: bool, optional
            If ``True``, no geometry will not be returned, by default ``False``.

        Returns
        -------
        geopandas.GeoDataFrame
            The query result as a ``geopandas.GeoDataFrame``.
        """
        if self.item is None or self.query_url is None:
            raise MissingItemError(["item"])

        cql_dict = {
            "json": cql_dict,
            "skipGeometry": str(skip_geometry).lower(),
        }
        return self._get_geodf(cql_dict)

    def __repr__(self) -> str:
        if self.item is None:
            return "\n".join(
                [
                    "Available Endpoints:",
                    "\n".join(f"'{k}': {v.description}" for k, v in self.endpoints.items()),
                ]
            )
        return "\n".join(
            [
                f"Item: '{self.item}'",
                f"Description: {self.endpoints[self.item].description}",
                f"Queryable Fields: {', '.join(self.endpoints[self.item].query_fields)}",
                f"Extent: {self.endpoints[self.item].extent}",
            ],
        )



"""Tests for PyNHD package."""

from __future__ import annotations

import io
import operator
import os
from pathlib import Path

import geopandas as gpd
import numpy as np
import pytest
from shapely import LineString, MultiPoint, Point, box

import pynhd
from pynhd import HP3D, NHD, NLDI, NHDPlusHR, PyGeoAPI, WaterData

is_ci = os.environ.get("GH_CI") == "true"
STA_ID = "01031500"
station_id = f"USGS-{STA_ID}"
site = "nwissite"
UM = "upstreamMain"
UT = "upstreamTributaries"


def assert_close(a: float, b: float) -> None:
    np.testing.assert_allclose(a, b, rtol=1e-3)


@pytest.fixture
def trib():
    comids = NLDI().navigate_byid(site, station_id, UT, "flowlines")
    return WaterData("nhdflowline_network").byid("comid", comids.nhdplus_comid.tolist())


def test_streamcat():
    nhd_area = pynhd.streamcat("fert", comids=13212248)
    assert_close(nhd_area["FERTWS"].item(), 14.358)
    nhd_area = pynhd.streamcat("inorgnwetdep2008", comids=23783629, lakes_only=True)
    assert_close(nhd_area["INORGNWETDEP2008WS"].item(), 1.7746)


def test_epa():
    data = pynhd.epa_nhd_catchments(9533477, "curve_number")
    assert_close(data["curve_number"].mean(axis=1), 75.576)

    data = pynhd.epa_nhd_catchments([9533477, 1440291], "comid_info")
    assert data["comid_info"].loc[1440291, "TOCOMID"] == 1439303


@pytest.mark.xfail(reason="Hydro is unstable.")
def test_nhd_xs_resample():
    main = NLDI().navigate_byid(site, station_id, UM, "flowlines")
    flw = NHD("flowline_mr").byids("COMID", main.nhdplus_comid.tolist()).to_crs(3857)
    main_nhd = pynhd.prepare_nhdplus(flw, 0, 0, 0, purge_non_dendritic=True)
    cs = pynhd.network_xsection(main_nhd, 2000, 1000)
    rs = pynhd.network_resample(main_nhd, 2000)
    assert len(cs) == 29
    assert len(rs) == 46


class TestPyGeoAPI:
    pygeoapi = PyGeoAPI()

    def test_flowtrace(self):
        coords = (1774209.63, 856381.68)
        gs = gpd.GeoDataFrame(
            {
                "direction": [
                    "none",
                ]
            },
            geometry=[Point(coords)],
            crs="ESRI:102003",
        )
        gdf = self.pygeoapi.flow_trace(coords, crs="ESRI:102003", direction="none")
        gdfb = pynhd.pygeoapi(gs, "flow_trace")
        assert gdf.comid.iloc[0] == gdfb.comid.iloc[0] == 22294818

    def test_splitcatchment(self):
        coords = (-73.82705, 43.29139)
        gs = gpd.GeoDataFrame(
            {
                "upstream": [
                    False,
                ]
            },
            geometry=[Point(coords)],
            crs=4326,
        )
        gdf = self.pygeoapi.split_catchment(coords, crs=4326, upstream=False)
        gdfb = pynhd.pygeoapi(gs, "split_catchment")
        assert gdf.catchmentID.iloc[0] == gdfb.catchmentID.iloc[0] == "22294818"

    @pytest.mark.xfail(reason="The xs endpoints of PyGeoAPI are not working.")
    def test_elevation_profile(self):
        line = LineString([(-103.801086, 40.26772), (-103.80097, 40.270568)])
        gs = gpd.GeoDataFrame(
            {
                "numpts": [
                    101,
                ],
                "dem_res": [
                    1,
                ],
            },
            geometry=[line],
            crs=4326,
        )
        gdf = self.pygeoapi.elevation_profile(line, numpts=101, dem_res=1, crs=4326)
        gdfb = pynhd.pygeoapi(gs, "elevation_profile")

        expected = 1299.8728
        assert_close(gdf.iloc[-1, 2], expected)
        assert_close(gdfb.iloc[-1, 2], expected)

    @pytest.mark.xfail(reason="The xs endpoints of PyGeoAPI are not working.")
    def test_endpoints_profile(self):
        coords = [(-103.801086, 40.26772), (-103.80097, 40.270568)]
        gs = gpd.GeoDataFrame(
            {
                "numpts": [
                    101,
                ],
                "dem_res": [
                    1,
                ],
            },
            geometry=[MultiPoint(coords)],
            crs=4326,
        )
        gdf = self.pygeoapi.endpoints_profile(coords, numpts=101, dem_res=1, crs=4326)
        gdfb = pynhd.pygeoapi(gs, "endpoints_profile")

        expected = 1299.8842
        assert_close(gdf.iloc[-1, 2], expected)
        assert_close(gdfb.iloc[-1, 2], expected)

    @pytest.mark.xfail(reason="The xs endpoints of PyGeoAPI are not working.")
    def test_cross_section(self):
        coords = (-103.80119, 40.2684)
        gs = gpd.GeoDataFrame(
            {
                "width": [
                    1000.0,
                ],
                "numpts": [
                    101,
                ],
            },
            geometry=[Point(coords)],
            crs=4326,
        )
        gdf = self.pygeoapi.cross_section(coords, width=1000.0, numpts=101, crs=4326)
        gdfb = pynhd.pygeoapi(gs, "cross_section")

        expected = 1301.482
        assert_close(gdf.iloc[-1, 2], expected)
        assert_close(gdfb.iloc[-1, 2], expected)


class TestNLDI:
    nldi: NLDI = NLDI()

    def test_navigate(self):
        stm = self.nldi.navigate_byid(site, station_id, UM, site)
        assert stm.shape[0] == 3

    def test_navigate_distance(self):
        st100 = self.nldi.navigate_byid(site, station_id, UM, site, distance=100)
        assert st100.shape[0] == 3

    def test_navigate_fsource(self):
        pp = self.nldi.navigate_byid(site, station_id, UT, "huc12pp")
        assert pp.shape[0] == 12

    def test_navigate_loc(self):
        wqp = self.nldi.navigate_byloc((-70, 44), UT, "wqp")
        assert wqp.comid.iloc[0] == "6710923"

    def test_comid_loc(self):
        station = self.nldi.getfeature_byid(site, station_id)
        lon = station.geometry.iloc[0].centroid.x
        lat = station.geometry.iloc[0].centroid.y
        comid = self.nldi.comid_byloc((lon, lat))
        assert station.comid.values[0] == comid.comid.values[0] == "1722317"

    def test_feature_loc(self):
        station = self.nldi.getfeature_byid(site, station_id)
        lon = round(station.geometry.iloc[0].centroid.x, 1)
        lat = round(station.geometry.iloc[0].centroid.y, 1)
        comid = self.nldi.feature_byloc((lon, lat))
        assert station.comid.values[0] == "1722317"
        assert comid.comid.values[0] == "1722211"

    def test_basin(self):
        eck4 = "+proj=eck4 +lon_0=0 +x_0=0 +y_0=0 +datum=WGS84 +units=km"
        basin = self.nldi.get_basins(STA_ID).to_crs(eck4)
        split = self.nldi.get_basins(STA_ID, split_catchment=True).to_crs(eck4)
        assert_close(split.area.values[0] - basin.area.values[0], -0.2489)

    def test_char(self):
        tot, prc = self.nldi.getcharacteristic_byid(
            "6710923", "local", char_ids="all", values_only=False
        )
        assert tot.CAT_BFI.values[0] == 57
        assert prc.CAT_BFI.values[0] == 0


class TestNHDAttrs:
    def test_meta_s3(self):
        meta = pynhd.nhdplus_attrs_s3()
        assert len(meta) == 14139

    def test_s3(self):
        attr = pynhd.nhdplus_attrs_s3("CAT_RECHG")
        assert_close(attr[attr.CAT_RECHG > 0].CAT_RECHG.mean(), 132.5881)

    def test_meta(self):
        meta = pynhd.nhdplus_attrs()
        assert len(meta) == 595

    def test_sb(self):
        attr = pynhd.nhdplus_attrs("BANKFULL")
        assert_close(attr[attr.BANKFULL_WIDTH > 0].BANKFULL_WIDTH.mean(), 13.6633)


class TestWaterData:
    def test_byid_flw(self, trib):
        assert_close(trib.lengthkm.sum(), 565.755)
        assert trib.shape[0] == 432

    def test_byid(self, trib):
        wd = WaterData("catchmentsp")
        ct = wd.byid("featureid", trib.comid.astype(str).to_list())
        assert_close(ct.areasqkm.sum(), 773.954)

    def test_bybox(self):
        wd = WaterData("wbd12")
        assert "wbd12" in wd.__repr__()
        wb_g = wd.bygeom(box(-118.72, 34.118, -118.31, 34.518), predicate="intersects", xy=True)
        wb_b = wd.bybox((-118.72, 34.118, -118.31, 34.518))
        assert_close(wb_b.areasqkm.sum(), wb_g.areasqkm.sum())

    def test_byfilter(self):
        crs = 3857
        wd = WaterData("huc12", crs)
        wb = wd.byfilter("huc12 LIKE '17030001%'", sort_attr="huc12")
        huc12 = wb[wb.huc12 == "170300010602"].geometry
        coords = (huc12.centroid.x, huc12.centroid.y)
        hucs = wd.bydistance(coords, 100, crs, sort_attr="huc12")
        assert wb.shape[0] == 52
        assert hucs.name[0] == "Upper Wenas River"


class TestGCX:
    def test_single(self):
        gcx = pynhd.GeoConnex("gages")
        gauge = gcx.byid("provider_id", "01031500")
        assert (gauge["nhdpv2_comid"] == 1722317).sum() == 1

    def test_multiple(self):
        gcx = pynhd.GeoConnex()
        gcx.item = "hu02"
        h2 = gcx.byid("huc2", "02")
        h3 = gcx.byid("huc2", "03")
        assert (h2["gnis_id"] == 2730132).sum() == (h3["gnis_id"] == 2730133).sum() == 1

    def test_many_features(self):
        gcx = pynhd.GeoConnex(max_nfeatures=10)
        gcx.item = "mainstems"
        ms = gcx.bygeometry((-69.77, 45.07, -69.31, 45.45))
        assert len(ms) == 20

    def test_cql(self):
        gcx = pynhd.GeoConnex("ua10")
        awa = gcx.bycql({"gt": [{"property": "awater10"}, 100e6]})
        assert len(awa) == 14


def test_3dhp():
    nhd3d = HP3D("flowline")
    flw = nhd3d.bygeom(Point(-89.441, 43.487), distance=10)
    assert flw.shape[0] == 1


def test_nhdphr():
    hr = NHDPlusHR("flowline")
    flwb = hr.bygeom((-69.77, 45.07, -69.31, 45.45))
    ids = ["103453218", "103454362", "103455178"]
    flwi = hr.byids("permanent_identifier", ids)
    ids_str = ", ".join([f"'{i}'" for i in ids])
    flwf = hr.bysql(f"permanent_identifier IN ({ids_str})")
    assert flwb.shape[0] == 3887
    assert sorted(flwi["permanent_identifier"]) == sorted(flwf["permanent_identifier"]) == ids


@pytest.mark.xfail(reason="Hydroshare is unstable.")
def test_nhdplus_vaa():
    fname = Path("nhdplus_vaa.parquet")
    vaa = pynhd.nhdplus_vaa(fname)
    fname.unlink()
    assert_close(vaa.slope.max(), 4.6)


class TestENHD:
    def test_wo_enhd_w_nan(self, trib: gpd.GeoDataFrame):
        attrs = pynhd.prepare_nhdplus(trib, 0, 0, 0, use_enhd_attrs=False, terminal2nan=True)
        assert attrs.tocomid.isna().sum() == 1

    def test_wo_enhd_wo_nan(self, trib: gpd.GeoDataFrame):
        attrs = pynhd.prepare_nhdplus(trib, 0, 0, 0, use_enhd_attrs=False, terminal2nan=False)
        assert attrs.tocomid.isna().sum() == 0

    def test_w_enhd_w_nan(self, trib: gpd.GeoDataFrame):
        attrs = pynhd.prepare_nhdplus(trib, 0, 0, 0, use_enhd_attrs=True, terminal2nan=True)
        assert attrs.tocomid.isna().sum() == 1


def test_toposort(trib):
    flw = pynhd.prepare_nhdplus(trib, 0, 0, 0, purge_non_dendritic=True)
    _, up_nodes, _ = pynhd.topoogical_sort(flw, id_col="comid", toid_col="tocomid")
    assert up_nodes[1721025] == [1720901]


def test_acc(trib):
    flw = pynhd.prepare_nhdplus(trib, 1, 1, 1, purge_non_dendritic=True)

    qsim = pynhd.vector_accumulation(
        flw[["comid", "tocomid", "lengthkm"]],
        operator.add,
        "lengthkm",
        ["lengthkm"],
    )
    flw = flw.merge(qsim, on="comid")
    diff = flw.arbolatesu - flw.acc_lengthkm
    assert_close(diff.abs().sum(), 439.451)


def test_h12pp():
    h12pp = pynhd.nhdplus_h12pp()
    assert len(h12pp) == 78249


def test_enhd_nx():
    g, m, s = pynhd.enhd_flowlines_nx()
    assert g.number_of_nodes() == len(m)
    assert s[0] == 8318775


def test_huc12_nx():
    g, m, s = pynhd.mainstem_huc12_nx()
    assert g.number_of_nodes() == len(m)
    assert s[0] == "150301040501"


def test_fcode():
    fcode = pynhd.nhd_fcode()
    assert fcode.loc[57100, "Feature Type"] == "DAM"


def test_show_versions():
    f = io.StringIO()
    pynhd.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()



"""Configuration for pytest."""

from __future__ import annotations

import pytest


@pytest.fixture(autouse=True)
def _add_standard_imports(doctest_namespace):
    """Add pynhd namespace for doctest."""
    import pynhd as nhd

    doctest_namespace["nhd"] = nhd



"""Access NLDI and WaterData databases."""

# pyright: reportGeneralTypeIssues=false
from __future__ import annotations

import contextlib
import io
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, cast

import cytoolz.curried as tlz
import geopandas as gpd
import numpy as np
import pandas as pd
import pyarrow.dataset as pds
from pyarrow import ArrowInvalid, fs

import async_retriever as ar
import pygeoogc as ogc
from pynhd.core import ScienceBase, get_parquet
from pynhd.exceptions import InputTypeError, InputValueError, ServiceError

if TYPE_CHECKING:
    from pyarrow.dataset import FileSystemDataset

__all__ = [
    "enhd_attrs",
    "nhdplus_vaa",
    "nhdplus_h12pp",
    "nhdplus_attrs",
    "nhdplus_attrs_s3",
    "nhd_fcode",
    "epa_nhd_catchments",
    "StreamCat",
    "streamcat",
]

NHDP_DTYPES = {
    "comid": "int32",
    "tocomid": "int32",
    "streamleve": "int8",
    "streamorde": "int8",
    "streamcalc": "int8",
    "fromnode": "int32",
    "tonode": "int32",
    "hydroseq": "int32",
    "levelpathi": "int32",
    "pathlength": "int32",
    "terminalpa": "int32",
    "arbolatesu": "int32",
    "divergence": "int8",
    "startflag": "int8",
    "terminalfl": "int8",
    "dnlevel": "int32",
    "thinnercod": "int8",
    "uplevelpat": "int32",
    "uphydroseq": "int32",
    "dnlevelpat": "int32",
    "dnminorhyd": "int32",
    "dndraincou": "int32",
    "dnhydroseq": "int32",
    "frommeas": "f8",
    "tomeas": "f8",
    "reachcode": str,
    "lengthkm": "f8",
    "fcode": "int32",
    "vpuin": "int8",
    "vpuout": "int8",
    "areasqkm": "f8",
    "totdasqkm": "f8",
    "divdasqkm": "f8",
    "totma": "f8",
    "wbareatype": str,
    "pathtimema": "f8",
    "slope": "f8",
    "slopelenkm": "f8",
    "ftype": str,
    "gnis_name": str,
    "gnis_id": "int32",
    "wbareacomi": "int32",
    "hwnodesqkm": "f8",
    "rpuid": str,
    "vpuid": str,
    "roughness": "f8",
}


def enhd_attrs(
    parquet_path: Path | str | None = None,
) -> pd.DataFrame:
    """Get updated NHDPlus attributes from ENHD V2.0.

    Notes
    -----
    This function downloads a 160 MB ``parquet`` file from
    `here <https://doi.org/10.5066/P976XCVT>`__.
    Although this dataframe does not include geometry, it can be
    linked to other geospatial NHDPlus dataframes through ComIDs.

    Parameters
    ----------
    parquet_path : str or pathlib.Pathlib.Path, optional
        Path to a file with ``.parquet`` extension for storing the file,
        defaults to ``./cache/enhd_attrs.parquet``.

    Returns
    -------
    pandas.DataFrame
        A dataframe that includes ComID-level attributes for
        2.7 million NHDPlus flowlines.
    """
    if parquet_path is None:
        output = get_parquet(Path("cache", "enhd_attrs.parquet"))
    else:
        output = get_parquet(parquet_path)

    url = ScienceBase.get_file_urls("63cb311ed34e06fef14f40a3").loc["enhd_nhdplusatts.parquet"].url
    url = cast("str", url)
    fname = ogc.streaming_download(url, file_extention=".parquet")
    if fname is None:
        msg = " ".join(
            (
                "The download was interrupted. Please try again.",
                "If the problem persists, please download the file manually",
                "from the following link and place it in the cache directory:",
                url,
            )
        )
        raise ServiceError(msg)

    enhd = pd.read_parquet(fname)
    dtype = {k: v for k, v in NHDP_DTYPES.items() if k in enhd.columns}
    enhd = enhd.astype(dtype, errors="ignore")
    if output.exists():
        output.unlink()
    enhd.to_parquet(output)
    return enhd


def nhdplus_vaa(
    parquet_path: Path | str | None = None,
) -> pd.DataFrame:
    """Get NHDPlus Value Added Attributes including roughness.

    Notes
    -----
    This function downloads a 245 MB ``parquet`` file from
    `here <https://www.hydroshare.org/resource/6092c8a62fac45be97a09bfd0b0bf726>`__.
    Although this dataframe does not include geometry, it can be linked
    to other geospatial NHDPlus dataframes through ComIDs.

    Parameters
    ----------
    parquet_path : str or pathlib.Pathlib.Path, optional
        Path to a file with ``.parquet`` extension for storing the file, defaults to
        ``./cache/nldplus_vaa.parquet``.

    Returns
    -------
    pandas.DataFrame
        A dataframe that includes ComID-level attributes for 2.7 million
        NHDPlus flowlines.
    """
    if parquet_path is None:
        output = get_parquet(Path("cache", "nldplus_vaa.parquet"))
    else:
        output = get_parquet(parquet_path)

    rid = "6092c8a62fac45be97a09bfd0b0bf726"
    fpath = "data/contents/nhdplusVAA.parquet"
    url = f"https://www.hydroshare.org/resource/{rid}/{fpath}"
    fname = ogc.streaming_download(url, file_extention=".parquet")
    if fname is None:
        msg = " ".join(
            (
                "The download was interrupted. Please try again.",
                "If the problem persists, please download the file manually",
                "from the following link and place it in the cache directory:",
                url,
            )
        )
        raise ServiceError(msg)

    vaa = pd.read_parquet(fname)
    dtype = {k: v for k, v in NHDP_DTYPES.items() if k in vaa.columns}
    vaa = vaa.astype(dtype, errors="ignore")
    if output.exists():
        output.unlink()
    vaa.to_parquet(output)
    return vaa


def _get_files(item: str) -> dict[str, tuple[str, str]]:
    """Get all the available zip files in an item."""
    url = "https://www.sciencebase.gov/catalog/item"
    payload = {"fields": "files,downloadUri", "format": "json"}
    resp = ar.retrieve_json([f"{url}/{item}"], [{"params": payload}])
    resp = cast("list[dict[str, Any]]", resp)
    files_url = zip(tlz.pluck("name", resp[0]["files"]), tlz.pluck("url", resp[0]["files"]))
    meta = list(tlz.pluck("metadataHtmlViewUri", resp[0]["files"], default=""))[-1]
    return {f.replace("_CONUS.zip", ""): (u, meta) for f, u in files_url if ".zip" in f}


def nhdplus_attrs(attr_name: str | None = None) -> pd.DataFrame:
    """Stage the NHDPlus Attributes database and save to nhdplus_attrs.parquet.

    Notes
    -----
    More info can be found `here <https://www.sciencebase.gov/catalog/item/5669a79ee4b08895842a1d47>`__.

    Parameters
    ----------
    attr_names : str , optional
        Name of NHDPlus attribute to return, defaults to None, i.e.,
        only return a metadata dataframe that includes the attribute names
        and their description and units.

    Returns
    -------
    pandas.DataFrame
        The staged data as a DataFrame.
    """
    sb = ScienceBase()
    r = sb.get_children("5669a79ee4b08895842a1d47")

    titles = tlz.pluck("title", r["items"])
    titles = tlz.concat(tlz.map(tlz.partial(re.findall, "Select(.*?)Attributes"), titles))
    titles = tlz.map(str.strip, titles)

    main_items = dict(zip(titles, tlz.pluck("id", r["items"])))

    files = {}
    soil = main_items.pop("Soil")
    for i, item in main_items.items():
        r = sb.get_children(item)

        titles = tlz.pluck("title", r["items"])
        titles = tlz.map(lambda s: s.split(":")[1].strip() if ":" in s else s, titles)

        child_items = dict(zip(titles, tlz.pluck("id", r["items"])))
        files[i] = {t: _get_files(c) for t, c in child_items.items()}

    r = sb.get_children(soil)
    titles = tlz.pluck("title", r["items"])
    titles = tlz.map(lambda s: s.split(":")[1].strip() if ":" in s else s, titles)

    child_items = dict(zip(titles, tlz.pluck("id", r["items"])))
    stat = child_items.pop("STATSGO Soil Characteristics")
    ssur = child_items.pop("SSURGO Soil Characteristics")
    files["Soil"] = {t: _get_files(c) for t, c in child_items.items()}

    r = sb.get_children(stat)
    titles = tlz.pluck("title", r["items"])
    titles = tlz.map(lambda s: s.split(":")[1].split(",")[1].strip(), titles)
    child_items = dict(zip(titles, tlz.pluck("id", r["items"])))
    files["STATSGO"] = {t: _get_files(c) for t, c in child_items.items()}

    r = sb.get_children(ssur)
    titles = tlz.pluck("title", r["items"])
    titles = tlz.map(lambda s: s.split(":")[1].strip(), titles)
    child_items = dict(zip(titles, tlz.pluck("id", r["items"])))
    files["SSURGO"] = {t: _get_files(c) for t, c in child_items.items()}

    chars = []
    types = {"CAT": "local", "TOT": "upstream_acc", "ACC": "div_routing"}
    for t, dd in files.items():
        for d, fd in dd.items():
            for f, u in fd.items():
                chars.append(
                    {
                        "name": f,
                        "type": types.get(f[-3:], "other"),
                        "theme": t,
                        "description": d,
                        "url": u[0],
                        "meta": u[1],
                    }
                )
    meta = pd.DataFrame(chars)
    if attr_name is None:
        return meta

    try:
        url = meta[meta.name == attr_name].url.values[0]
    except IndexError as ex:
        raise InputValueError("name", meta.name.unique().tolist()) from ex

    fname = ogc.streaming_download(url, file_extention="zip")
    if fname is None:
        msg = " ".join(
            (
                "The download was interrupted. Please try again.",
                "If the problem persists, please download the file manually",
                "from the following link and place it in the cache directory:",
                url,
            )
        )
        raise ServiceError(msg)
    return pd.read_csv(fname, engine="pyarrow")


def nhdplus_attrs_s3(
    attr_names: str | list[str] | None = None, nodata: bool = False
) -> pd.DataFrame:
    """Access NHDPlus V2.1 derived attributes over CONUS.

    Notes
    -----
    More info can be found `here <https://www.sciencebase.gov/catalog/item/5669a79ee4b08895842a1d47>`__.

    Parameters
    ----------
    attr_names : str or list of str, optional
        Names of NHDPlus attribute(s) to return, defaults to None, i.e.,
        only return a metadata dataframe that includes the attribute names
        and their description and units.
    nodata : bool
        Whether to include NODATA percentages, default is False.

    Returns
    -------
    pandas.DataFrame
        A dataframe of requested NHDPlus attributes.
    """
    bucket = "prod-is-usgs-sb-prod-publish"
    meta_url = "/".join(
        (f"https://{bucket}.s3.amazonaws.com", "5669a79ee4b08895842a1d47/metadata_table.tsv")
    )
    resp = ar.retrieve_text([meta_url])
    meta = pd.read_csv(io.StringIO(resp[0]), delimiter="\t")
    if attr_names is None:
        return meta

    urls = meta.set_index("ID").datasetURL.str.rsplit("/", n=1).str[-1].to_dict()

    attr_names = attr_names if isinstance(attr_names, (list, tuple)) else [attr_names]
    if any(a not in urls for a in attr_names):
        raise InputValueError("attr_names", list(urls))
    ids = tlz.merge_with(list, ({urls[c]: c} for c in attr_names))

    s3 = fs.S3FileSystem(anonymous=True, region="us-west-2")
    get_dataset = tlz.partial(pds.dataset, filesystem=s3, format="parquet")
    datasets = (get_dataset(f"{bucket}/{i}/{i}_{c[0][:3].lower()}.parquet") for i, c in ids.items())

    def to_pandas(ds: FileSystemDataset, cols: list[str], nodata: bool) -> pd.DataFrame:
        """Convert a pyarrow dataset to a pandas dataframe."""
        cols = ["COMID", *cols]
        if nodata:
            ds_columns = ds.schema.names
            with contextlib.suppress(StopIteration):
                cols.append(next(c for c in ds_columns if "NODATA" in c))
        return ds.to_table(columns=cols).to_pandas().set_index("COMID")

    return pd.concat((to_pandas(d, c, nodata) for d, c in zip(datasets, ids.values())), axis=1)


def nhdplus_h12pp(gpkg_path: Path | str | None = None) -> pd.DataFrame:
    """Access HUC12 Pour Points for NHDPlus V2.1 L48 (CONUS).

    Notes
    -----
    More info can be found
    `here <https://www.sciencebase.gov/catalog/item/60cb5edfd34e86b938a373f4>`__.

    Parameters
    ----------
    gpkg_path : str or pathlib.Pathlib.Path, optional
        Path to the geopackage file, defaults to None, i.e., download
        the file to the cache directory as ``102020wbd_outlets.gpkg``.

    Returns
    -------
    geopandas.GeoDataFrame
        A geodataframe of HUC12 pour points.
    """
    url = ScienceBase.get_file_urls("60cb5edfd34e86b938a373f4").loc["102020wbd_outlets.gpkg"].url
    url = cast("str", url)
    gpkg_path = Path("cache", "102020wbd_outlets.gpkg") if gpkg_path is None else Path(gpkg_path)
    gpkg_path = ogc.streaming_download(url, fnames=gpkg_path)
    if gpkg_path is None:
        msg = " ".join(
            (
                "The download was interrupted. Please try again.",
                "If the problem persists, please download the file manually",
                "from the following link and place it in the cache directory:",
                url,
            )
        )
        raise ServiceError(msg)
    h12pp = gpd.read_file(gpkg_path, engine="pyogrio")
    h12pp = h12pp[
        ["COMID", "REACHCODE", "REACH_meas", "offset", "HUC12", "LevelPathI", "geometry"]
    ].copy()
    h12pp[["COMID", "LevelPathI"]] = h12pp[["COMID", "LevelPathI"]].astype("uint32")
    h12pp = cast("gpd.GeoDataFrame", h12pp)
    return h12pp


def nhd_fcode() -> pd.DataFrame:
    """Get all the NHDPlus FCodes."""
    url = "/".join(
        [
            "https://gist.githubusercontent.com/cheginit",
            "8f9730f75b2b9918f492a236c77c180c/raw/nhd_fcode.json",
        ]
    )
    return pd.read_json(url)


def epa_nhd_catchments(
    comids: int | str | list[int | str],
    feature: Literal["curve_number", "comid_info"],
) -> dict[str, pd.DataFrame]:
    """Get NHDPlus catchment-scale data from EPA's HMS REST API.

    Notes
    -----
    For more information about curve number please refer to the project's
    webpage on the EPA's
    `website <https://cfpub.epa.gov/si/si_public_record_Report.cfm?Lab=CEMM&dirEntryId=351307>`__.

    Parameters
    ----------
    comids : int or list of int
        ComID(s) of NHDPlus catchments.
    feature : str
        The feature of interest. Available options are:

        - ``curve_number``: 16-day average Curve Number.
        - ``comid_info``: ComID information.

    Returns
    -------
    dict of pandas.DataFrame or geopandas.GeoDataFrame
        A dict of the requested dataframes. A ``comid_info`` dataframe is
        always returned.

    Examples
    --------
    >>> import pynhd
    >>> data = pynhd.epa_nhd_catchments(9533477, "curve_number")
    >>> data["curve_number"].mean(axis=1).item()
    75.576
    """
    feature_names = {
        "curve_number": "cn",
        "comid_info": "",
    }
    if feature not in feature_names:
        raise InputValueError("feature", list(feature_names))

    clist = comids if isinstance(comids, (list, tuple)) else [comids]
    f_kwd = {feature_names[feature]: "true"} if feature != "comid_info" else {}
    urls, kwds = zip(
        *[
            (
                "https://qed.epa.gov/hms/rest/api/info/catchment",
                {"params": {**f_kwd, "comid": comid}},
            )
            for comid in clist
        ]
    )
    urls = cast("list[str]", urls)
    kwds = cast("list[dict[str, Any]]", kwds)
    resp = ar.retrieve_json(urls, kwds)
    resp = cast("list[dict[str, Any]]", resp)
    info = pd.DataFrame.from_dict(
        {i: pd.Series(r["metadata"]) for i, r in zip(clist, resp)}, orient="index"
    )
    for c in info:
        info[c] = pd.to_numeric(info[c], errors="coerce")

    if feature == "curve_number":
        data = pd.DataFrame.from_dict(
            {i: r["curve_number"] for i, r in zip(clist, resp)}, orient="index", dtype="f8"
        )
        return {"comid_info": info, "curve_number": data}

    return {"comid_info": info}


class StreamCat:
    """Get StreamCat API's properties.

    Parameters
    ----------
    lakes_only : bool, optional
        If ``True``, only return metrics for lakes and their associated catchments
        from the LakeCat dataset.

    Attributes
    ----------
    base_url : str
        The base URL of the API.
    valid_names : list of str
        The valid names of the metrics.
    alt_names : dict of str
        The alternative names of some metrics.
    valid_regions : list of str
        The valid hydro regions.
    valid_states : pandas.DataFrame
        The valid two letter states' abbreviations.
    valid_counties : pandas.DataFrame
        The valid counties' FIPS codes.
    valid_aois : list of str
        The valid types of areas of interest.
    metrics_df : pandas.DataFrame
        The metrics' metadata such as description and units.
    valid_years : dict
        A dictionary of the valid years for annual metrics.
    """

    def __init__(self, lakes_only: bool = False) -> None:
        self.lakes_only = lakes_only
        if self.lakes_only:
            self.base_url = "https://java.epa.gov/StreamCAT/LakeCat/metrics"
        else:
            self.base_url = "https://java.epa.gov/StreamCAT/metrics"
        resp = ar.retrieve_json([self.base_url])
        resp = cast("list[dict[str, Any]]", resp)
        params = resp[0]["parameters"]

        self.valid_names = cast("list[str]", params["name"]["options"])
        self.alt_names = {
            n.replace("_", ""): n for n in self.valid_names if n.split("_")[-1].isdigit()
        }
        self.alt_names.update(
            {
                "precip2008": "precip08",
                "precip2009": "precip09",
                "tmean2008": "tmean08",
                "tmean2009": "tmean09",
            }
        )
        self.valid_regions = params["region"]["options"]
        self.valid_states = pd.DataFrame.from_dict(params["state"]["options"], orient="index")
        self.valid_counties = pd.DataFrame.from_dict(params["county"]["options"], orient="index")
        self.valid_aois = params["areaOfInterest"]["options"]
        self.valid_slopes = tlz.merge_with(
            list,
            (
                {s[:-9]: f"slp{slp}" for s in self.valid_names if f"slp{slp}" in s}
                for slp in (10, 20)
            ),
        )

        url_vars = f"{self.base_url}/variable_info.csv"
        names = pd.read_csv(io.BytesIO(ar.retrieve_binary([url_vars])[0]), encoding="latin1")
        names["METRIC_NAME"] = names["METRIC_NAME"].str.replace(r"\[AOI\]|Slp[12]0", "", regex=True)
        names["SLOPE"] = [
            ", ".join(self.valid_slopes.get(m.replace("[Year]", "").lower(), []))
            for m in names.METRIC_NAME
        ]
        names.loc[names["SLOPE"] == "", "SLOPE"] = np.nan
        slp = ~names["SLOPE"].isna()
        names.loc[slp, "METRIC_NAME"] = [f"{n}[Slope]" for n in names.loc[slp, "METRIC_NAME"]]
        names = (
            names.drop_duplicates("METRIC_NAME").reset_index(drop=True).drop(columns="WEBTOOL_NAME")
        )
        self.metrics_df = names

        years = names.set_index("METRIC_NAME").YEAR.dropna()
        self.valid_years = {
            str(v): list(range(*(int(y) for y in yrs.split("-"))))
            if "-" in yrs
            else [int(y) for y in yrs.split(",")]
            for v, yrs in years.items()
        }


class StreamCatValidator(StreamCat):
    def __init__(self, lakes_only: bool = False) -> None:
        super().__init__(lakes_only)

    def validate(
        self,
        name: list[str] | None = None,
        region: list[str] | None = None,
        state: list[str] | None = None,
        county: list[str] | None = None,
        aoi: list[str] | None = None,
    ) -> None:
        """Validate input parameters."""
        if name and not set(name).issubset(self.valid_names):
            raise InputValueError("metric_names", self.valid_names)

        if region and not set(region).issubset(self.valid_regions):
            raise InputValueError("regions", self.valid_regions)

        if state and not set(state).issubset(self.valid_states.index):
            raise InputValueError("states", self.valid_states.index.to_list())

        if county and not set(county).issubset(self.valid_counties.index):
            raise InputValueError("counties", self.valid_counties.index.to_list())

        if aoi and not set(aoi).issubset(self.valid_aois):
            raise InputValueError("metric_areas", self.valid_aois)

    def id_kwds(
        self,
        comids: int | list[int] | None = None,
        regions: str | list[str] | None = None,
        states: str | list[str] | None = None,
        counties: str | list[str] | None = None,
        conus: bool = False,
    ) -> dict[str, str]:
        """Get the keyword arguments for the API's ID parameters."""
        n_args = sum(i is not None for i in (comids, regions, states, counties))
        n_args += int(conus)
        if n_args != 1:
            msg = "Exactly one of comids, regions, states, counties, or conus must be passed"
            raise ValueError(msg)

        params = {}
        if comids is not None:
            try:
                comid_arr = np.array(
                    [comids] if isinstance(comids, (str, int)) else comids, dtype=int
                )
                params["comid"] = ",".join(comid_arr.astype(str))
            except (ValueError, TypeError) as ex:
                raise InputTypeError("comids", "int or digists (str) or list of them") from ex
        elif regions is not None:
            ids = [regions] if isinstance(regions, str) else regions
            self.validate(region=ids)
            params["region"] = ",".join(ids)
        elif states is not None:
            ids = [states.upper()] if isinstance(states, str) else [s.upper() for s in states]
            self.validate(state=ids)
            params["state"] = ",".join(ids)
        elif conus:
            params["conus"] = "true"
        return params


def streamcat(
    metric_names: str | list[str] | None = None,
    metric_areas: str | list[str] | None = None,
    comids: int | list[int] | None = None,
    regions: str | list[str] | None = None,
    states: str | list[str] | None = None,
    counties: str | list[str] | None = None,
    conus: bool = False,
    percent_full: bool = False,
    area_sqkm: bool = False,
    lakes_only: bool = False,
) -> pd.DataFrame:
    """Get various metrics for NHDPlusV2 catchments from EPA's StreamCat.

    Notes
    -----
    For more information about the service check its webpage
    at https://www.epa.gov/national-aquatic-resource-surveys/streamcat-dataset.


    Parameters
    ----------
    metric_names : str or list of str, optional
        Metric name(s) to retrieve. There are 567 metrics available.
        to get a full list check out :meth:`StreamCat.valid_names`.
        To get a description of each metric, check out
        :meth:`StreamCat.metrics_df`. Some metrics require year and/or slope
        to be specified, which have ``[Year]`` and/or ``[Slope]`` in their name.
        For convenience all these variables and their years/slopes are converted
        to a dict that can be accessed via :meth:`StreamCat.valid_years` and
        :meth:`StreamCat.valid_slopes`. Defaults to ``None``, which will return
        a dataframe of the metrics metadata.
    metric_areas : str or list of str, optional
        Areas to return the metrics for, defaults to ``None``, i.e. all areas.
        Valid options are: ``catchment``, ``watershed``, ``riparian_catchment``,
        ``riparian_watershed``, ``other``.
    comids : int or list of int, optional
        NHDPlus COMID(s), defaults to ``None``. Either ``comids``, ``regions``,
        ``states``, ``counties``, or ``conus`` must be passed. They are
        mutually exclusive.
    regions : str or list of str, optional
        Hydro region(s) to retrieve metrics for, defaults to ``None``. For a
        full list of valid regions check out :meth:`StreamCat.valid_regions`
        Either ``comids``, ``regions``, ``states``, ``counties``, or ``conus``
        must be passed. They are mutually exclusive.
    states : str or list of str, optional
        Two letter state abbreviation(s) to retrieve metrics for, defaults to
        ``None``. For a full list of valid states check out
        :meth:`StreamCat.valid_states` Either ``comids``, ``regions``,
        ``states``, ``counties``, or ``conus`` must be passed. They are
        mutually exclusive.
    counties : str or list of str, optional
        County FIPS codes(s) to retrieve metrics for, defaults to ``None``. For
        a full list of valid county codes check out :meth:`StreamCat.valid_counties`
        Either ``comids``, ``regions``, ``states``, ``counties``, or ``conus`` must
        be passed. They are mutually exclusive.
    conus : bool, optional
        If ``True``, ``metric_names`` of all NHDPlus COMIDs are retrieved,
        defaults ``False``. Either ``comids``, ``regions``,
        ``states``, ``counties``, or ``conus`` must be passed. They are mutually
        exclusive.
    percent_full : bool, optional
        If ``True``, return the percent of each area of interest covered by
        the metric.
    area_sqkm : bool, optional
        If ``True``, return the area in square kilometers.
    lakes_only : bool, optional
        If ``True``, only return metrics for lakes and their associated catchments
        from the LakeCat dataset.

    Returns
    -------
    pandas.DataFrame
        A dataframe with the requested metrics.
    """
    if metric_names is None:
        return StreamCat().metrics_df
    sc = StreamCatValidator(lakes_only)
    names = [metric_names] if isinstance(metric_names, str) else metric_names
    names = [sc.alt_names.get(s.lower(), s.lower()) for s in names]
    sc.validate(name=names)
    params = {"name": ",".join(n for n in names)}

    if metric_areas:
        aoi = [metric_areas] if isinstance(metric_areas, str) else metric_areas
        sc.validate(aoi=aoi)
        params["areaOfInterest"] = ",".join(aoi)

    ids = sc.id_kwds(comids, regions, states, counties, conus)
    if ids:
        params.update(ids)

    if percent_full:
        params["showPctFull"] = "true"

    if area_sqkm:
        params["showAreaSqKm"] = "true"

    params_str = "&".join(f"{k}={v}" for k, v in params.items())
    if len(params_str) < 7000:
        resp = ar.retrieve_text([f"{sc.base_url}?{params_str}"])
    else:
        resp = ar.retrieve_text([sc.base_url], [{"data": params_str}], request_method="post")

    try:
        return pd.read_csv(io.StringIO(resp[0]), engine="pyarrow")
    except ArrowInvalid as ex:
        raise ServiceError(resp[0]) from ex
